#-----------------------------------------------------------------------------
# Swarm   v4
# 6/27/04
# Copyright (c) 2004 David G. Drumright (ockham)
#-----------------------------------------------------------------------------

from Tkinter import *
root = Tk()
import tkFileDialog
import poser
import math
import random
import os
scene = poser.Scene()
BasePathUncut=poser.AppLocation()
BasePath=BasePathUncut[:(-len('poser.exe'))]
MyPath=os.path.normpath(BasePath+'/Runtime/Libraries/props/*.PP2')

#-----------------------------------------------------------------------------
# Next section is TK
#------------------------------------------------------------

class App:
    def __init__(self, master, textMessage):
        self.master = master
        master.title("Swarm")
        self.PartNameList=[] # Needed to hold info between select and go
        self.PreserveShapeVar=IntVar()
        self.PreserveShapeVar.set(0)

        self.StatusEntry = Entry(self.master,width=40)
        self.StatusEntry.grid(row=0,column=0)
        self.SelectPartButton = Button(self.master, text="Select particle", command=self.HandleSelect)
        self.SelectPartButton.grid(row=1, column=0)

        self.ListFrame=Frame(self.master,borderwidth=2,relief=RIDGE)
        self.ListFrame.grid(row=2,column=0,sticky=W+E)
        self.SliderFrame=Frame(self.master,borderwidth=2,relief=RIDGE)
        self.SliderFrame.grid(row=3,column=0,sticky=W+E)
        self.ActionFrame=Frame(self.master,borderwidth=2,relief=RIDGE)
        self.ActionFrame.grid(row=4,column=0,sticky=W+E)
        # Go button.
        self.GoButton = Button(self.master, text="Go", command=self.HandleGo)
        self.GoButton.grid(row=5, column=0)

        # Source and Dest material listboxes.
        self.SourceLabel=Label(self.ListFrame,text="Source materials")
        self.SourceLabel.grid(row=0,column=1)
        self.SourceScroll  = Scrollbar(self.ListFrame, orient=VERTICAL)
        self.SourceScroll.grid( row=1, column=0,sticky=N+S+E)
        self.SourceList = Listbox(self.ListFrame, height=5, selectmode=SINGLE,exportselection=0,yscrollcommand = self.SourceScroll.set)
        self.SourceList.grid( row=1, column=1)
        self.SourceScroll["command"] = self.SourceList.yview

        self.DestLabel=Label(self.ListFrame,text="Particle materials")
        self.DestLabel.grid(row=0,column=3)
        self.DestScroll  = Scrollbar(self.ListFrame, orient=VERTICAL)
        self.DestScroll.grid( row=1, column=2,sticky=N+S+E)
        self.DestList = Listbox(self.ListFrame, height=5, selectmode=MULTIPLE,exportselection=0,yscrollcommand = self.DestScroll.set)
        self.DestList.grid( row=1, column=3)
        self.DestScroll["command"] = self.DestList.yview

        self.CountLabel=Label(self.SliderFrame, text="Number:")
        self.CountLabel.grid(row=0,column=0,sticky=E)
        self.CountSlider=Scale(self.SliderFrame, from_ = 1, to = 100,orient=HORIZONTAL)
        # Note that this starts and defaults at 1.
        self.CountSlider.grid(row=0, column=1)

        self.OrigSizeLabel=Label(self.SliderFrame, text="Size as percent of base prop:")
        self.OrigSizeLabel.grid( row=1,column=0,sticky=E)
        self.OrigSizeSlider=Scale(self.SliderFrame, from_ = 1, to = 100,orient=HORIZONTAL)
        self.OrigSizeSlider.grid(row=1,column=1)
        self.OrigSizeSlider.set(10) # Default of 10
        self.PreserveCheck=Checkbutton(self.SliderFrame,text="Preserve Shape?",variable=self.PreserveShapeVar)
        self.PreserveCheck.grid(row=1,column=2)

        self.RandSizeLabel=Label(self.SliderFrame, text="Randomness of size:")
        self.RandSizeLabel.grid( row=2,column=0,sticky=E)
        self.RandSizeSlider=Scale(self.SliderFrame, from_ = 0, to = 100,orient=HORIZONTAL)
        self.RandSizeSlider.grid(row=2,column=1)

        self.RandDirLabel=Label(self.SliderFrame, text="Randomness of orientation:")
        self.RandDirLabel.grid( row=3,column=0,sticky=E)
        self.RandDirSlider=Scale(self.SliderFrame, from_ = 0, to = 100,orient=HORIZONTAL)
        self.RandDirSlider.grid(row=3,column=1)

        self.SpeedLabel=Label(self.ActionFrame, text="Speed of action:")
        self.SpeedLabel.grid( row=0,column=0,sticky=E)
        self.SpeedSlider=Scale(self.ActionFrame, from_ = 0, to = 100,orient=HORIZONTAL)
        self.SpeedSlider.grid(row=0,column=1)

        self.StartLabel=Label(self.ActionFrame, text="Start frame:")
        self.StartLabel.grid( row=1,column=0,sticky=E)
        self.StartEntry=Entry(self.ActionFrame, width=6)
        self.StartEntry.grid(row=1,column=1)

        self.EndLabel=Label(self.ActionFrame, text="End frame:")
        self.EndLabel.grid( row=2,column=0,sticky=E)
        self.EndEntry=Entry(self.ActionFrame, width=6)
        self.EndEntry.grid(row=2,column=1)

        # Set the default values for frame entries
        self.StartEntry.insert(0,str(1))
        self.EndEntry.insert(0,str(scene.NumFrames()))

        # Set up the Base first.
        # The base must be the selected item.
        self.BaseItem = scene.CurrentActor()
        # If it's part of a figure, get the figure as well.
        self.BaseFig = self.BaseItem.ItsFigure()
        if (not self.BaseItem.IsBodyPart()) and (not self.BaseItem.IsProp()):
            self.ShowStatus("Base prop not a good type.  Cancel and restart.")
        else:
            self.ShowStatus("First select the 'particle' prop.")

        # Set up close action
        self.master.protocol('WM_DELETE_WINDOW', self.HandleCancel)

    # - - - - - - - - - - - - - - - - - -

    def ShowStatus(self,S): # Just saving repetition
        self.StatusEntry.delete(0,END)
        self.StatusEntry.insert(0,S)
        self.StatusEntry.update_idletasks()

    # - - - - - - - - - - - - - - - - - -

    def HandleSelect(self): # Get and analyze the Particle.
        # Pop up file box to get a Particle.
        self.FileName=tkFileDialog.askopenfilename(initialfile=MyPath, filetypes=[("Props", "PP2")] )
        if not self.FileName:
            self.ShowStatus("No file selected. Try again.")
            return

        self.ShowStatus("Working.......")

        self.PartNameList=[] # Holds just the names.  The only reason we need this is
        # because we load one example first to get orig size, then load the
        # others in a separate loop.

        # Get the sliders we need here.
        Count = self.CountSlider.get()
        OrigSize = float(self.OrigSizeSlider.get()) / 100.0
        RandSize = float(self.RandSizeSlider.get()) / 100.0
        RandDir  = float(self.RandDirSlider.get())  / 100.0



        # Fetch one copy of the particle, so we can get its
        # bounding box in orig state.  This copy will become the
        # first instance; the rest (if Count>1) will just take
        # the data.
        try:
            scene.LoadLibraryProp(os.path.normpath(self.FileName))
        except:
            self.ShowStatus("Prop not usable. Try again.")
            return

        # The new prop will be CurrentActor at this moment.
        self.SamplePart = scene.CurrentActor()
        self.PartNameList.append(self.SamplePart.Name())

        # Get the material info from the source and dest, and fill the lists.
        if self.BaseFig:
            self.BaseMatlList = self.BaseFig.Materials()
        else:
            self.BaseMatlList = self.BaseItem.Materials()
        for OneMatl in self.BaseMatlList:
            self.SourceList.insert(END,OneMatl.Name())
        self.PartMatlList = self.SamplePart.Materials()
        for OneMatl in self.PartMatlList:
            self.DestList.insert(END,OneMatl.Name())


        # The actual value settings of the sample particle will be done along with the others
        # in HandleGo.
        self.ShowStatus("Now set values and materials, then hit Go.")


    # - - - - - - - - - - - - - - - - - -

    def HandleGo(self):
        self.SetupAll()
        self.Swarm()
        self.ShowStatus("Done.")

    # - - - - - - - - - - - - - - - - - -

    def SetupAll(self):
        self.ShowStatus("Working....")
        # Get the sliders we need here.
        Count = self.CountSlider.get()
        OrigSize = float(self.OrigSizeSlider.get()) / 100.0
        RandSize = float(self.RandSizeSlider.get()) / 100.0
        RandDir  = float(self.RandDirSlider.get())  / 100.0

        # Pick up the names from the materials lists. If no sel,
        # use first in each.
        Sels=self.SourceList.curselection()
        if Sels:
            BaseMatlName = self.SourceList.get(int(Sels[0]))
        else:
            BaseMatlName = self.SourceList.get(0)
        BaseMatl = self.BaseItem.Material(BaseMatlName)

        Sels=self.DestList.curselection()
        if Sels:
            PartMatlName = self.DestList.get(int(Sels[0]))
        else:
            PartMatlName = self.DestList.get(0)
        # We don't assign PartMatl here, because it's separate
        # for each of the generated particles.

        # To shorten the loop, get direct values now:
        BaseTexCol = BaseMatl.TextureColor()
        BaseSpeCol = BaseMatl.SpecularColor()
        BaseDifCol = BaseMatl.DiffuseColor()
        BaseAmbCol = BaseMatl.AmbientColor()
        BaseRefCol = BaseMatl.ReflectionColor()
        BaseRefVal = BaseMatl.ReflectionStrength()
        BaseBumVal = BaseMatl.BumpStrength()
        BaseTrnMax = BaseMatl.TransparencyMax()
        BaseTrnMin = BaseMatl.TransparencyMin()
        BaseTrnFal = BaseMatl.Ns()
        BaseTexMap = BaseMatl.TextureMapFileName()
        BaseRefMap = BaseMatl.ReflectionMapFileName()
        BaseTrnMap = BaseMatl.TransparencyMapFileName()
        BaseBumMap = BaseMatl.BumpMapFileName()

        scene.DrawAll() # Always needed when using WVs

        # Get the bounding box of the base thing.  (This needs to be
        # more precise, but for a start the bound will do.)
        # Note that this is done by WorldVerts, so you don't need
        # to have the object in default position; but the rectangle
        # may be much larger than the real thing in oblique cases.
        MaxX=-10000.0
        MaxY=-10000.0
        MaxZ=-10000.0
        MinX= 10000.0
        MinY= 10000.0
        MinZ= 10000.0
        BaseGeom=self.BaseItem.Geometry()
        for Vindex in range(BaseGeom.NumVertices()):
            OneVert=BaseGeom.WorldVertex(Vindex)
            X=OneVert.X()
            Y=OneVert.Y()
            Z=OneVert.Z()
            if   X < MinX: MinX = X
            elif X > MaxX: MaxX = X
            if   Y < MinY: MinY = Y
            elif Y > MaxY: MaxY = Y
            if   Z < MinZ: MinZ = Z
            elif Z > MaxZ: MaxZ = Z

        # For easier figuring, find the center and widths:
        self.CenterX = (MaxX + MinX) / 2.0  # Midpoint
        self.CenterY = (MaxY + MinY) / 2.0
        self.CenterZ = (MaxZ + MinZ) / 2.0
        BaseRadiusX  = abs(MaxX - MinX) / 2.0  # Half the width
        BaseRadiusY  = abs(MaxY - MinY) / 2.0
        BaseRadiusZ  = abs(MaxZ - MinZ) / 2.0
        # Preserve the Base bounding box as class vars,
        # for use in the action part.
        self.maxX=MaxX
        self.minX=MinX
        self.maxY=MaxY
        self.minY=MinY
        self.maxZ=MaxZ
        self.minZ=MinZ

        # Now get the bounding box of the part in its loaded state.
        MaxX=-10000.0
        MaxY=-10000.0
        MaxZ=-10000.0
        MinX= 10000.0
        MinY= 10000.0
        MinZ= 10000.0
        PartGeom=self.SamplePart.Geometry()
        for Vindex in range(PartGeom.NumVertices()):
            OneVert=PartGeom.WorldVertex(Vindex)
            X=OneVert.X()
            Y=OneVert.Y()
            Z=OneVert.Z()
            if   X < MinX: MinX = X
            elif X > MaxX: MaxX = X
            if   Y < MinY: MinY = Y
            elif Y > MaxY: MaxY = Y
            if   Z < MinZ: MinZ = Z
            elif Z > MaxZ: MaxZ = Z

        # In this case we don't need the midpoint
        PartRadiusX  = abs(MaxX - MinX) / 2.0  # Half the width
        PartRadiusY  = abs(MaxY - MinY) / 2.0
        PartRadiusZ  = abs(MaxZ - MinZ) / 2.0

        # Form the starting scale, using the OrigSize slider along with the
        # ratio of Part to Base.
        OrigXscale = OrigSize * PartRadiusX / BaseRadiusX
        OrigYscale = OrigSize * PartRadiusY / BaseRadiusY
        OrigZscale = OrigSize * PartRadiusZ / BaseRadiusZ
        # Create the rest of the particles:
        for i in range(Count-1): # Count-1 because we already have the first
            # Create a copy of the Part.
            try:
                scene.LoadLibraryProp(os.path.normpath(self.FileName))
            except:
                self.ShowStatus("Prop not usable. Try again.")
                return
            # The new prop will be currentactor at this moment.
            Part = scene.CurrentActor()
            self.PartNameList.append(Part.Name())

        # Set the original locations and angles.  The original locs are
        # uniformly distributed with slight random.  To get a sort-of uniform
        # dist, we begin by assuming that the number of particles in each
        # direction is the cube root of the total count.  Then we look at
        # the proportions of the bounding box.
        # We multiply the X count by X/Y, and divide the Y count by X/Y.
        # We multiply the Z count by Z/Y, and divide the Y count by Z/Y.
        # If this gives less than 1 in any direction, we take 1.
        # Then we place particles using these counts, with the intervals
        # figured in the obvious way.

        CubeRoot = math.pow(float(Count),(1.0/3.0))
        Xcount=CubeRoot * (BaseRadiusX/BaseRadiusY)
        Ycount=CubeRoot / (BaseRadiusX/BaseRadiusY)
        Zcount=CubeRoot * (BaseRadiusZ/BaseRadiusY)
        Ycount=CubeRoot / (BaseRadiusZ/BaseRadiusY)
        if Xcount < 1.0: Xcount=1.0
        if Ycount < 1.0: Ycount=1.0
        if Zcount < 1.0: Zcount=1.0
        Xinterval = BaseRadiusX * 2.0 / (Xcount + 1.0)
        Yinterval = BaseRadiusY * 2.0 / (Ycount + 1.0)
        Zinterval = BaseRadiusZ * 2.0 / (Zcount + 1.0)

        Xordinal=0
        Yordinal=0
        Zordinal=0


        for PartName in self.PartNameList: # List includes the sample
            Part = scene.Actor(PartName)
            # For convenience, get the necessary parms of all piecees:
            PartXrotParm =   Part.ParameterByCode(poser.kParmCodeXROT)
            PartYrotParm =   Part.ParameterByCode(poser.kParmCodeYROT)
            PartZrotParm =   Part.ParameterByCode(poser.kParmCodeZROT)

            PartXtranParm =  Part.ParameterByCode(poser.kParmCodeXTRAN)
            PartYtranParm =  Part.ParameterByCode(poser.kParmCodeYTRAN)
            PartZtranParm =  Part.ParameterByCode(poser.kParmCodeZTRAN)

            PartXscaleParm = Part.ParameterByCode(poser.kParmCodeXSCALE)
            PartYscaleParm = Part.ParameterByCode(poser.kParmCodeYSCALE)
            PartZscaleParm = Part.ParameterByCode(poser.kParmCodeZSCALE)
            PartAscaleParm = Part.ParameterByCode(poser.kParmCodeASCALE)

            # Set the Y and X rot with desired degree of randomness. (No point
            # in setting all three; gimbal lock and allstufflikethatthere.)
            PartYrotParm.SetValue(PartYrotParm.Value()+360.0 * RandDir * random.random())
            PartXrotParm.SetValue(PartXrotParm.Value()+360.0 * RandDir * random.random())

            # Set the scale with desired degree of randomness.
            if self.PreserveShapeVar:
                R = RandSize * (random.random()-0.5)
                if R < 0.01: R = 0.01
                PartAscaleParm.SetValue(R)
            else:
                R = OrigXscale + (RandSize * (random.random()-0.5))
                if R < 0.01: R = 0.01
                PartXscaleParm.SetValue(R)
                R = OrigYscale + (RandSize * (random.random()-0.5))
                if R < 0.01: R = 0.01
                PartYscaleParm.SetValue(R)
                R = OrigZscale + (RandSize * (random.random()-0.5))
                if R < 0.01: R = 0.01
                PartZscaleParm.SetValue(R)

            # Now place the locations.  Start from Center vals, and move
            # X,Y,Z by a random proportion of the Radius vals.  The random is
            # first transformed from [0,1.0) to [-1.0,1.0)

            #RandXtran = self.CenterX + (BaseRadiusX *  (random.random() * 2.0 - 1.0))
            #RandYtran = self.CenterY + (BaseRadiusY *  (random.random() * 2.0 - 1.0))
            #RandZtran = self.CenterZ + (BaseRadiusZ *  (random.random() * 2.0 - 1.0))
            # Use the Intervals. For now no random; after this looks right, add bit of random.
            PartXtranParm.SetValue(self.CenterX - BaseRadiusX + (Xordinal * Xinterval) + (Xinterval*(random.random()-0.5)))
            PartYtranParm.SetValue(self.CenterY - BaseRadiusY + (Yordinal * Yinterval) + (Yinterval*(random.random()-0.5)))
            PartZtranParm.SetValue(self.CenterZ - BaseRadiusZ + (Zordinal * Zinterval) + (Zinterval*(random.random()-0.5)))

            # Parent the particle to the baseitem.
            Part.SetParent(self.BaseItem,0,1)

            # Assign the BaseItem's chosen material to the part's chosen material.
            PartMatl = Part.Material(PartMatlName)
            # Arbitrarily take the first one as default.
            (R,G,B)=BaseTexCol
            PartMatl.SetTextureColor(R,G,B)
            (R,G,B)=BaseSpeCol
            PartMatl.SetSpecularColor(R,G,B)
            (R,G,B)=BaseDifCol
            PartMatl.SetDiffuseColor(R,G,B)
            (R,G,B)=BaseAmbCol
            PartMatl.SetAmbientColor(R,G,B)
            (R,G,B)=BaseRefCol
            PartMatl.SetReflectionColor(R,G,B)
            PartMatl.SetReflectionStrength(BaseRefVal)
            PartMatl.SetTransparencyMax(BaseTrnMax)
            PartMatl.SetTransparencyMin(BaseTrnMin)
            PartMatl.SetNs(BaseTrnFal)
            PartMatl.SetTextureMapFileName(BaseTexMap)
            PartMatl.SetBumpStrength(BaseBumVal)
            PartMatl.SetReflectionMapFileName(BaseRefMap)
            PartMatl.SetTransparencyMapFileName(BaseTrnMap)
            PartMatl.SetBumpMapFileName(BaseBumMap)
            Part.Memorize()
            # For the sake of doing the Intervals.
            # Move fastest on X, slowest on Z.
            Xordinal=Xordinal+1
            if Xordinal>=Xcount:
                Xordinal=0
                Yordinal=Yordinal+1
                if Yordinal>=Ycount:
                    Yordinal=0
                    Zordinal=Zordinal+1


    # - - - - - - - - - - - - - - - - - -

    def Swarm(self): # Acts only if speed has been set
        Speed = float(self.SpeedSlider.get()) / 100.0
        if Speed < 0.0001: return  # Allow for slight twitches of slider

        # Fetch start and end:
        StartFrame=1
        EndFrame=scene.NumFrames()
        s = self.StartEntry.get()
        if s:
            StartFrame = int(s)
        s = self.EndEntry.get()
        if s:
            EndFrame = int(s)
        if (EndFrame - StartFrame) < 2: return

        # Now rig up the increment. Applying something like the Nyquist notion,
        # I think the fastest possible speed (100) should take a particle across
        # the smallest dimension of the bounding box in one frame.
        BaseXscale = self.BaseItem.ParameterByCode(poser.kParmCodeXSCALE).Value()
        BaseYscale = self.BaseItem.ParameterByCode(poser.kParmCodeYSCALE).Value()
        BaseZscale = self.BaseItem.ParameterByCode(poser.kParmCodeZSCALE).Value()
        BaseAscale = self.BaseItem.ParameterByCode(poser.kParmCodeASCALE).Value()
        NetXscale = BaseXscale * BaseAscale
        NetYscale = BaseYscale * BaseAscale
        NetZscale = BaseZscale * BaseAscale
        # Modify the max/mins to be within the parented box:
        MaxX = (self.maxX - self.CenterX) / NetXscale
        MinX = (self.minX - self.CenterX) / NetXscale
        MaxY = (self.maxY - self.CenterY) / NetYscale
        MinY = (self.minY - self.CenterY) / NetYscale
        MaxZ = (self.maxZ - self.CenterZ) / NetZscale
        MinZ = (self.minZ - self.CenterZ) / NetZscale

        dx = MaxX - MinX
        dy = MaxY - MinY
        dz = MaxZ - MinZ
        if   (dx <= dy) and (dx <= dz): Narrowest=dx
        elif (dy <= dx) and (dy <= dz): Narrowest=dy
        elif (dz <= dx) and (dz <= dy): Narrowest=dz

        # Narrowest is the increment when speed=100, so:
        Increment = Narrowest * Speed

        # From startframe to endframe, move each particle randomly,
        # with increment determined by Speed var.  Keep them within
        # the bounding box.
        for F in range(StartFrame+1,EndFrame): # Leave the first frame as it was.
            scene.SetFrame(F-1) # Change to 1-based at last moment
            for PartName in self.PartNameList: # List includes the sample
                Part = scene.Actor(PartName)
                PartXtran =  Part.ParameterByCode(poser.kParmCodeXTRAN)
                PartYtran =  Part.ParameterByCode(poser.kParmCodeYTRAN)
                PartZtran =  Part.ParameterByCode(poser.kParmCodeZTRAN)
                NewXVal = PartXtran.Value() + ((random.random() * 2.0 - 1.0) * Increment)
                NewYVal = PartYtran.Value() + ((random.random() * 2.0 - 1.0) * Increment)
                NewZVal = PartZtran.Value() + ((random.random() * 2.0 - 1.0) * Increment)
                if   NewXVal > MaxX: NewXVal = MaxX
                elif NewXVal < MinX: NewXVal = MinX
                if   NewYVal > MaxY: NewYVal = MaxY
                elif NewYVal < MinY: NewYVal = MinY
                if   NewZVal > MaxZ: NewZVal = MaxZ
                elif NewZVal < MinZ: NewZVal = MinZ
                PartXtran.SetValue(NewXVal)
                PartYtran.SetValue(NewYVal)
                PartZtran.SetValue(NewZVal)

    # - - - - - - - - - - - - - - - - - -

    def HandleCancel(self):
        self.master.destroy()

    # - - - - - - - - - - - - - - - - - -

    def Update(self):
        if scene: scene.ProcessSomeEvents(1)
        root.lift()
        root.after(250, self.Update)

#------------------------------------------------------------
# Activate the loop
#------------------------------------------------------------

app = App(root, '')
app.Update()
root.mainloop()

#------------------------------------------------------------
# End TK loop.
#-----------------------------------------------------------------------------
